from __future__ import unicode_literals

from .config import DataConfig, file_read_bytes, file_write_bytes, file_write_text, file_move, file_delete, KeyPtr
from .cryptography import SaltHandle, CryptHandle
from .create_key import create_key

__all__ = [
    "DataConfig", "CryptHandle", "SaltHandle",
    "create_key", "file_read_bytes", "file_write_bytes", "file_write_text", "file_move",
    "file_delete", "KeyPtr"
]


from __future__ import unicode_literals

from threading import Thread, Lock
from sqlalchemy.engine import Engine
from sqlalchemy.exc import SQLAlchemyError
from pyodbc import Error as PYODBCError
from pandas import DataFrame
from csv import writer as csv_writer, QUOTE_ALL
from types import FunctionType, BuiltinFunctionType, MethodType, BuiltinMethodType

import logging
import os

log = logging.getLogger(__name__)


class SQLCursor(Thread):
    """
    A SQLCursor class to handle SQLEngine's cursor's when executing queries or retrieving server tables
    -   Class is threaded thus multiple cursors can be executing in a queue at the same time
    """

    __cursor = None
    __engine_class = None
    __cursor_action = None

    def __init__(self, engine_type, engine, spid, keep_engine_alive=True, engine_class=None, action=None,
                 action_params=None):
        """
        Creates a SQLCursor class instance that handles engine cursor operations

        :param engine: engine that is generated from SQLEngine's instance class
        :param engine_class: SQLEngine class instance
        :param action: [execute, tables]
        :param action_params: Action parameters according to execute or tables command below
        """

        if engine_type not in ['alchemy', 'pyodbc']:
            raise ValueError("'engine_type' %r is not alchemy or pyodbc" % engine_type)
        if action and action not in ('execute', 'tables'):
            raise ValueError("'action' is not execute or tables")

        self.__engine_type = engine_type
        self.__engine = engine
        self.__spid = spid
        self.engine_class = engine_class
        self.__raw_engine = None
        self.__keep_engine_alive = keep_engine_alive

        if engine_type == 'alchemy':
            self.__raw_engine = self.__engine.raw_connection()
            engine = self.__raw_engine
        else:
            engine = self.__engine

        self.__cursor = engine.cursor()
        if action:
            self.__cursor_action = [action, action_params]
        self.__execute_errors = None
        self.__execute_results = None
        self.__is_pending = Lock()
        self.__is_closing = Lock()
        self.cursor_id = sum(map(ord, str(os.urandom(100))))
        super(SQLCursor, self).__init__()

    @property
    def cursor_action(self):
        """
        :return: Returns cursor action that was performed
        """

        return self.__cursor_action

    @property
    def results(self):
        """
        :return: Return one or more datasets retreived from execution of sql command
        """

        return self.__execute_results

    @property
    def errors(self):
        """
        :return: Returns one or more errors that occurred from execution of sql command
        """

        return self.__execute_errors

    @property
    def is_pending(self):
        """
        :return: Returns True/False when cursor is actively processing an action
        """

        return self.__is_pending.locked()

    @property
    def engine_class(self):
        """
        :return: Returns SQLEngine class instance that may be tied to this class
        """

        return self.__engine_class

    @engine_class.setter
    def engine_class(self, engine_class):
        from ..sql.engine import SQLEngineClass

        if not isinstance(engine_class, (SQLEngineClass, type(None))):
            raise ValueError("'engine_class' %r is not an instance of SQLEngineClass" % engine_class)

        self.__engine_class = engine_class

    def run(self):
        """
        Run operator that is executed upon creation of this SQLCursor class instance
        """

        if self.__cursor_action:
            function, params = self.__cursor_action

            if function == 'execute':
                if 'query_str' not in params.keys():
                    raise ValueError("'query_str' was not provided as action_param when class ran")
                if 'execute' not in params.keys():
                    params['execute'] = None

                if 'handlers' in params.keys():
                    for handler, buffer, csv_path, delimiter, quotechar in params['handlers']:
                        self.execute(query_str=params['query_str'], execute=params['execute'], handler=handler,
                                     buffer=buffer, csv_path=csv_path, delimiter=delimiter, quotechar=quotechar)
                else:
                    self.execute(query_str=params['query_str'], execute=params['execute'])
            elif function == 'tables':
                self.tables()
            elif function is not None:
                raise ValueError("'function' is an invalid function command")

    def close(self, write_log=True):
        """
        Closes cursor comm, pop SQLCursor instance class out of SQLEngine's cursor queue and add to results list in
        SQLEngine's class instance

        :param write_log: (Optional) [True/False]
        """

        with self.__is_closing:
            if self.__cursor and hasattr(self.__cursor, 'cancel'):
                try:
                    if write_log:
                        if self.__engine_class:
                            log.debug('SQL Connection (%s): Canceling & Closing SQL cursor %s transaction',
                                      self.__engine_class.get_engine_id, self.cursor_id)
                        else:
                            log.debug('Canceling & Closing SQL cursor %s transaction', self.cursor_id)

                    self.__cursor.cancel()
                except:
                    pass
                finally:
                    self.__close()

    def rollback(self, write_log=True):
        """
        Rollsback cursor comm and executes close() operations within SQLCursor class instance

        :param write_log: (Optional) [True/False]
        """

        with self.__is_closing:
            if self.__cursor and hasattr(self.__cursor, 'rollback'):
                try:
                    if write_log:
                        if self.__engine_class:
                            log.debug('SQL Connection (%s): Rollback & Close SQL cursor %s transaction',
                                      self.__engine_class.get_engine_id, self.cursor_id)
                        else:
                            log.debug('Rollback & Close SQL cursor %s transaction', self.cursor_id)

                    self.__cursor.rollback()
                except:
                    pass
                finally:
                    self.__close()

    def commit(self, write_log=True):
        """
        Commit cursor comm and executes close() operations within SQLCursor class instance

        :param write_log: (Optional) [True/False]
        """

        try:
            with self.__is_closing:
                if self.__cursor and hasattr(self.__cursor, 'commit'):
                    if write_log:
                        if self.__engine_class:
                            log.debug('SQL Connection (%s): Commit & Close SQL cursor %s transaction',
                                      self.__engine_class.get_engine_id, self.cursor_id)
                        else:
                            log.debug('Commit & Close SQL cursor %s transaction', self.cursor_id)

                    self.__cursor.commit()
                    self.__close()
        except:
            self.rollback()
            pass

    def __close(self):
        from .engine import close_engine

        if self.__cursor and hasattr(self.__cursor, 'close'):
            try:
                self.__cursor.close()
            except:
                pass

        if not self.__keep_engine_alive:
            close_engine(self.__engine)

        if self.__engine_class:
            if self.__execute_results or self.__execute_errors:
                self.__engine_class.add_cursor_result(self)

            self.__engine_class.rem_cursor(self)

        self.__engine = None
        self.__raw_engine = None
        self.__cursor = None

    def tables(self):
        """
        Requests sql connection to pull [table_type, table_cat, table_schema, table_name] and store it in a dataframe.
        Dataframe is appended to the results attribute of this class
        """

        if self.__cursor:
            try:
                if not self.__cursor_action:
                    self.__cursor_action = ['tables', None]

                with self.__is_pending:
                    log.debug("Retreiving tables on SPID %s" % self.__spid)
                    self.__execute_results = list()
                    tables = [[t.table_type, t.table_cat, t.table_schem, t.table_name] for t in self.__cursor.tables()]

                    if tables:
                        self.__execute_results.append(DataFrame(tables, columns=['Table_Type', 'Table_Cat',
                                                                                 'Table_Schema', 'Table_Name']))
                    else:
                        self.__execute_results.append(DataFrame())

                    self.__close()
            except SQLAlchemyError as e:
                self.__execute_errors = [e.code, e.__dict__['orig']]
                self.rollback()
            except PYODBCError as e:
                self.__execute_errors = [e.args[0], e.args[1]]
                self.rollback()
            except (AttributeError, Exception) as e:
                self.__execute_errors = [type(e).__name__, str(e)]
                self.rollback()
        else:
            raise Exception("Cursor is closed. Cannot pull tables")

    def execute(self, query_str, execute=False, handler=None, buffer=1000, csv_path=None, csv_replace=False,
                delimiter=',', quotechar='"', quoting=QUOTE_ALL):
        """
        Requests sql connection to execute or query a sql query string. Execution of TSQL queries will follow
        commit and rollback commands when necessary. Results are appended to the results and/or errors attributes

        :param query_str: SQL query string
        :param execute: (Optional) [True/False] Default is False
        :param handler: (Optional) Function or Class Function handler
        :param buffer: (Optional) [handler only] Number of lines per chunk to store in dataframe
        :param csv_path: (Optional) [handler only] File path where data is stored at for csv file
        :param csv_replace: (Optional) [True/False] Replace csv file if exists
        :param delimiter: (Optional) [handler only] Data seperator to delimit columns
        :param quotechar: (Optional) [handler only] Quote Character to wrap values with
        """

        if handler:
            if not isinstance(handler, (FunctionType, BuiltinFunctionType, MethodType, BuiltinMethodType)):
                raise ValueError("'handler' %r is not a function" % handler)
            if not isinstance(buffer, int):
                raise eError("'buffer' %r is not a int" % buffer)
            if csv_path and not isinstance(csv_path, str):
                raise ValueError("'csv_path' %r is not a str" % csv_path)
            if not isinstance(delimiter, str):
                raise ValueError("'delimiter' %r is not a str" % delimiter)
            if not isinstance(quotechar, str):
                raise ValueError("'quotechar' %r is not a str" % quotechar)

            buffer = int(buffer)

            if buffer < 1:
                raise ValueError("'buffer' %r value is zero or negative" % buffer)
            if csv_path and not os.path.exists(os.path.dirname(csv_path)):
                raise ValueError("'csv_path' %r is not a valid directory path" % os.path.dirname(csv_path))
            if csv_path and not csv_path.endswith('.csv'):
                raise ValueError("'csv_path' %r is does not have a .csv extension" % os.path.basename(csv_path))

            if csv_path and csv_replace and os.path.exists(csv_path):
                try:
                    os.remove(csv_path)
                except:
                    pass

            if csv_path and os.path.exists(csv_path):
                raise ValueError("'csv_path' %r file exists already" % csv_path)

        if self.__cursor:
            if not isinstance(query_str, str):
                raise ValueError("'query_str' %r is not a String" % query_str)

            try:
                if not self.__cursor_action:
                    self.__cursor_action = ['execute', dict(query_str=query_str, execute=execute)]

                with self.__is_pending:
                    log.debug("Executing on SPID {0} query [{1}]".format(self.__spid, query_str))
                    self.__execute_results = list()
                    result = self.__cursor.execute(query_str)

                    if handler:
                        self.__stream_dataset(result, handler, buffer, csv_path, delimiter, quotechar, quoting)
                    else:
                        self.__store_dataset(result)

                    while result.nextset():
                        if handler:
                            self.__stream_dataset(result, handler, buffer, csv_path, delimiter, quotechar, quoting)
                        else:
                            self.__store_dataset(result)

                    if execute:
                        self.commit()
                    else:
                        self.__close()
            except SQLAlchemyError as e:
                self.__execute_errors = [e.code, e.__dict__['orig']]
                self.rollback()
            except PYODBCError as e:
                self.__execute_errors = [e.args[0], e.args[1]]
                self.rollback()
            except (AttributeError, Exception) as e:
                from traceback import format_exc
                print(format_exc())
                self.__execute_errors = [type(e).__name__, str(e)]
                self.rollback()
        else:
            raise Exception("Cursor is closed. Cannot execute query")

    def __store_dataset(self, dataset):
        try:
            data = [tuple(t) for t in dataset.fetchall()]
            cols = [column[0] for column in dataset.description]
            self.__execute_results.append(DataFrame(data, columns=cols))
        except:
            pass

    def __stream_dataset(self, dataset, handler, buffer, csv_path, delimiter, quotechar, quoting):
        cols = [column[0] for column in dataset.description]
        buffer_list = list()
        row_num = 0

        for row in dataset:
            buffer_list.append(tuple(row))

            if buffer <= len(buffer_list):
                self.__handle_buffer(row_num, cols, buffer_list, handler, csv_path, delimiter, quotechar, quoting)
                buffer_list.clear()

            row_num += 1

        self.__handle_buffer(row_num - 1, cols, buffer_list, handler, csv_path, delimiter, quotechar, quoting)

    @staticmethod
    def __handle_buffer(row_num, cols, buffer_list, handler, csv_path, delimiter, quotechar, quoting):
        if buffer_list:
            df = DataFrame(buffer_list, columns=cols)

            if csv_path:
                hreturn = handler(df, row_num - len(buffer_list) + 1, row_num)

                if hreturn and isinstance(hreturn, DataFrame):
                    df = hreturn

                if row_num - len(buffer_list) + 1 == 0:
                    mode = 'w'
                    header = True
                else:
                    mode = 'a'
                    header = False

                df.to_csv(path_or_buf=csv_path, sep=delimiter, quotechar=quotechar, mode=mode, index=False,
                          header=header, quoting=quoting)
            else:
                handler(df, row_num - len(buffer_list) + 1, row_num)


class EngineCursor(Thread):
    """
    Engine Cursor class for SQLAlchemy connections. Primary purpose of this class is to utilize dataframe to_upload()
    """

    __engine = None
    __engine_class = None

    def __init__(self, alch_engine, spid, keep_engine_alive=True, engine_class=None, action=None, action_params=None):
        """
        Creates a EngineCursor class instance and run an action with parameters upon class creation

        :param alch_engine: SQLAlchemy engine
        :param engine_class: (Optional) SQLEngine instance class
        :param action: (Optional) [upload]
        :param action_params: (Optional) upload_df() command parameters
        """

        if not isinstance(alch_engine, Engine):
            raise ValueError("'alch_engine' %r is not an Alchemy Engine instance" % alch_engine)

        if action and action != 'upload_df':
            raise ValueError("'action' is not upload_df")

        self.engine_class = engine_class
        self.__engine = alch_engine
        self.__spid = spid
        self.__keep_engine_alive = keep_engine_alive
        self.__cursor_action = [action, action_params]
        self.__is_pending = Lock()
        self.__is_closing = Lock()
        self.__errors = None
        self.cursor_id = sum(map(ord, str(os.urandom(100))))
        super(EngineCursor, self).__init__()

    @property
    def cursor_action(self):
        """
        :return: Returns cursor action that was performed
        """

        return self.__cursor_action

    @property
    def results(self):
        """
        :return: Return one or more datasets retreived from execution of sql command
        """

        return None

    @property
    def errors(self):
        """
        :return: Returns one or more errors that occurred from execution of sql command
        """

        return self.__errors

    @property
    def is_pending(self):
        """
        :return: Returns True/False when cursor is actively processing an action
        """

        return self.__is_pending.locked()

    @property
    def engine_class(self):
        return self.__engine_class

    @engine_class.setter
    def engine_class(self, engine_class):
        from ..sql.engine import SQLEngineClass

        if not isinstance(engine_class, (SQLEngineClass, type(None))):
            raise ValueError("'engine_class' %r is not an instance of SQLEngineClass" % engine_class)

        self.__engine_class = engine_class

    def run(self):
        """
        Run operator that is executed upon creation of this SQLCursor class instance
        """

        if self.__cursor_action:
            function, params = self.__cursor_action

            if function == 'upload_df':
                if 'dataframe' not in params.keys():
                    raise ValueError("'dataframe' was not provided as action_param when class ran")
                if 'table_name' not in params.keys():
                    raise ValueError("'table_name' was not provided as action_param when class ran")
                if 'table_schema' not in params.keys():
                    params['table_schema'] = None
                if 'if_exists' not in params.keys():
                    params['if_exists'] = None
                if 'index' not in params.keys():
                    params['index'] = None
                if 'index_label' not in params.keys():
                    params['index_label'] = None

                self.upload_df(dataframe=params['dataframe'], table_name=params['table_name'],
                               table_schema=params['table_schema'], if_exists=params['if_exists'],
                               index=params['index'], index_label=params['index_label'])
            elif function is not None:
                raise ValueError("'function' %s is an invalid function command" % function)

    def close(self):
        """
        Closes cursor comm, pop SQLCursor instance class out of SQLEngine's cursor queue and add to results list in
        SQLEngine's class instance
        """

        with self.__is_closing:
            if self.__engine:
                from .engine import close_engine

                try:
                    if hasattr(self.__engine, 'cancel'):
                        self.__engine.cancel()

                    if hasattr(self.__engine, 'rollback'):
                        self.__engine.rollback()
                except SQLAlchemyError as e:
                    log.error(e.code, e.__dict__['orig'])
                    pass

                if not self.__keep_engine_alive:
                    close_engine(self.__engine)

            self.__engine = None

    def upload_df(self, dataframe, table_name, table_schema=None, if_exists='append', index=True, index_label='ID'):
        """
        Uploads a pandas DataFrame to a sql table

        :param dataframe: pandas Dataframe
        :param table_name: SQL table name
        :param table_schema: (Optional) SQL table schema
        :param if_exists: (Optional) [append, replace]
        :param index: (Optional) [True, False] Default is True
        :param index_label: (Optional) Default is ID
        """

        if self.__engine:
            if not isinstance(dataframe, DataFrame):
                raise ValueError("'dataframe' %r is not an pandas Dataframe instance" % dataframe)
            if not isinstance(table_name, str):
                raise ValueError("'table_name' %r is not a String" % table_name)
            if not isinstance(table_schema, (str, type(None))):
                raise ValueError("'table_schema' %r is not a String" % table_schema)

            if not self.__cursor_action:
                params = dict(dataframe=dataframe, table_name=table_name, table_schema=table_schema,
                              if_exists=if_exists, index=index, index_label=index_label)
                self.__cursor_action = ['upload_df', params]

            with self.__is_pending:
                log.debug("SQL Upload on SPID {0} to [{1}]".format(self.__spid, '{0}.{1}'.join(table_schema,
                                                                                               table_name)))

                try:
                    dataframe.to_sql(
                        table_name,
                        self.__engine,
                        schema=table_schema,
                        if_exists=if_exists,
                        index=index,
                        index_label=index_label,
                        chunksize=1000
                    )
                except SQLAlchemyError as e:
                    self.__errors = [e.code, e.__dict__['orig']]
                    self.close()
                except (AttributeError, Exception) as e:
                    self.__errors = [type(e).__name__, str(e)]
                    self.close()
                else:
                    self.__close()
        else:
            raise ValueError('Engine is closed. Unable to upload dataframe')

    def __close(self):
        if self.__engine_class:
            if self.__errors:
                self.__engine_class.add_cursor_result(self)

            self.__engine_class.rem_cursor(self)
        else:
            self.__engine = None



from ..filehandler import FileHandler
from xml.etree.ElementTree import parse as xml_parse
from pandas import DataFrame
from os.path import exists, join
from datetime import datetime
from portalocker import Lock
from xml.sax.saxutils import escape


class XMLReader(FileHandler):
    def __init__(self, file_path, xmlns_rs=None, dict_var=None):
        super().__init__(file_path=file_path, xmlns_rs=xmlns_rs, dict_var=dict_var)

    def parse(self):
        if self.streams:
            for handler, buffer in self.streams:
                self.__parse(handler=handler, buffer=buffer)
        else:
            return self.__parse()

    def __parse(self, handler=None, buffer=None):
        from ..filehandler import is_an_xml

        if is_an_xml(self.file_path):
            tree = xml_parse(self.file_path)
            root = tree.getroot()
            xml_objs = root.findall(self.xmlns_rs)

            if isinstance(self.dict_var, dict):
                for item in xml_objs:
                    self.dict_var = self.__parse_element(item, self.dict_var)

                return self.dict_var
            else:
                self.__parse_xml_root(xml_objs, handler, buffer)

    @staticmethod
    def __parse_xml_root(xml_objs, handler=None, buffer=None):
        xml_dicts = list()
        row_num = 0
        header = None

        for xml_obj in xml_objs:
            xml_dicts.append(self.__parse_element(xml_obj))

            if not header:
                header = converted_row

            if handler and buffer <= len(xml_dicts):
                handler(self.__parse_df(xml_dicts), row_num - len(xml_dicts) + 1)
                xml_dicts.clear()
                xml_dicts.append(header)

            row_num += 1

        if handler:
            row_num -= 1
            handler(self.__parse_df(xml_dicts), row_num - len(xml_dicts) + 1)
        else:
            return self.__parse_df(xml_dicts)

    @staticmethod
    def __parse_df(xml_dicts):
        if xml_dicts:
            df = DataFrame(xml_dicts)
            df.applymap(lambda x: x.strip() if isinstance(x, str) else x)
        else:
            df = DataFrame()

        return df

    def __parse_element(self, element, parsed=None):
        if not parsed:
            parsed = dict()

        if element.keys():
            for key in element.keys():
                if key not in parsed:
                    parsed[key] = element.attrib.get(key)

                if element.text and element.tag not in parsed:
                    parsed[element.tag] = element.text

        elif element.text and element.tag not in parsed:
            parsed[element.tag] = element.text

        for child in list(element):
            self.__parse_element(child, parsed)

        return parsed


class XMLWriter(FileHandler):
    def __init__(self, file_dir, file_name, df=None):
        super().__init__(file_dir=file_dir, file_name=file_name, df=df)

    def write(self):
        xml_file = self.__xml_file()

        with Lock(xml_file, 'w') as xmlfile:
            xmlfile.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            xmlfile.write('<records>\n')
            xmlfile.write(
                '\n'.join(dataframe.apply(self.__xml_encode, axis=1))
            )
            xmlfile.write('\n</records>')

    def __xml_file(self):
        from ..filehandler import unique_id

        xml_file = join(self.file_dir, self.file_name)
        date = datetime.now()

        while exists(xml_file):
            uid = unique_id()
            xml_file = join(self.file_dir, '%s-%s-%s_%s.xml' % (date.year, date.month, date.day, uid))

        return xml_file

    def __xml_encode(self, row):
        xmlitem = ['  <record>']

        for field in row.index:
            if row[field]:
                xmlitem.append('    <var var_name="{0}">{1}</var>'.format(field, self.__handle_illegals(row[field])))

        xmlitem.append('  </record>')

        return '\n'.join(xmlitem)

    @staticmethod
    def __handle_illegals(data):
        return escape(str(data).replace('"', '&quot;').replace("'", '&apos;').replace("’", '&apos;')
                      .replace(chr(160), ' '))



from __future__ import unicode_literals

from datetime import datetime
from tkinter.messagebox import showerror, askokcancel
from multiprocessing import Pipe
from tkinter import *
from threading import Lock, Thread

import os
import logging
import sys
import string


class LogHandle(object):
    __file_dir = None
    __base_name = None
    __gui_console = None
    __write_pipe = None
    __read_pipe = None
    __pipe_thread = None

    """
    Logging that becomes easier and more uniformed
    """

    def __init__(self, file_dir=None, base_name=None):
        """
        Logging settings not forgotten!

        :param file_dir: File directory of where log files will be saved
        :param base_name: A base filename that log files will be saved as
        """

        self.file_dir = file_dir
        self.base_name = base_name
        self.__printable = set(string.printable)

    @property
    def write_pipe(self):
        if not self.__write_pipe:
            if self.__read_pipe:
                self.__read_pipe.close()

            self.__write_pipe, self.__read_pipe = Pipe()

        if not self.__pipe_thread:
            self.__pipe_thread = Thread(target=self.__pipe_listen)
            self.__pipe_thread.daemon = True
            self.__pipe_thread.start()

        return self.__write_pipe

    @property
    def file_dir(self):
        return self.__file_dir

    @file_dir.setter
    def file_dir(self, file_dir):
        if not isinstance(file_dir, (str, type(None))):
            raise Exception("'file_dir' is missing")
        if file_dir and not os.path.exists(file_dir):
            os.makedirs(file_dir)

        self.__file_dir = file_dir

    @property
    def base_name(self):
        return self.__base_name

    @base_name.setter
    def base_name(self, base_name):
        if not isinstance(base_name, (str, type(None))):
            raise Exception("'base_name' is missing")

        self.__base_name = base_name

    @property
    def gc(self):
        return self.__gui_console

    @gc.setter
    def gc(self, gc):
        self.__gui_console = gc

    @staticmethod
    def __clean_handlers():
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

    def __pipe_listen(self):
        while True:
            args = self.__read_pipe.recv()

            while isinstance(args, list) and not args[0]:
                args.remove(args[0])

            if isinstance(args, list) and len(args) == 1 and isinstance(args[0], list):
                args = args[0]

            if isinstance(args, dict) and 'command' in args.keys():
                command = args['command']
                del args['command']

                if command == 'gui_console':
                    self.__gui_console_set(**args)
                elif command == 'write_to_log':
                    self.__write_to_log(**args)
            elif isinstance(args, list) and len(args) > 1:
                command = args[0]

                if command == 'gui_console':
                    self.__gui_console_set(*args)
                elif command == 'write_to_log':
                    self.__write_to_log(*args)

    def gui_console(self, title=None, destroy=False, turn_off=False, gui_obj=None):
        if gui_obj and hasattr(gui_obj, "print_gui") and hasattr(gui_obj, "log_setup"):
            self.__gui_console = gui_obj
            self.__gui_console.log_setup(log_class=self)
        else:
            self.__gui_console_set(title=title, destroy=destroy, turn_off=turn_off)

    def write_to_log(self, message, action='info', print_only=False):
        self.__write_to_log(message=message, action=action, print_only=print_only)

    def __gui_console_set(self, title=None, destroy=False, turn_off=False, *args, **kwargs):
        if self.__gui_console and destroy:
            self.__gui_console.destroy()
            self.__gui_console = None
        elif self.__gui_console and turn_off:
            self.__gui_console = None
        else:
            self.__gui_console = GUIConsole(parent=self, title=title)

    def __write_to_log(self, message, action='info', print_only=False, *args, **kwargs):
        """
        Logging becomes more organized!

        :param message: Message that you would like to write to log
        :param action: What kind of action would you like the entry to be marked as?
            (debug, info, warning, error, critical)
        :param print_only: (True/False) print line only and not log into file
        """

        if not print_only and self.__base_name and self.__file_dir:
            if not message:
                raise Exception("'message' is missing")

            self.__clean_handlers()
            filepath = os.path.join(self.__file_dir,
                                    "{0}_{1}_Log.txt".format(datetime.now().__format__("%Y%m%d"), self.__base_name))

            logging.basicConfig(filename=filepath,
                                level=logging.DEBUG, format=' %(asctime)s - %(levelname)s - %(message)s')

        message = ''.join(filter(lambda x: x in self.__printable, message))

        if self.__gui_console:
            self.__gui_console.print_gui('{0} - {1} - {2}'.format(datetime.now().__format__("%Y%m%d %I:%M:%S %p"),
                                                                  action.upper(), message))
        else:
            print('{0} - {1} - {2}'.format(datetime.now(), action.upper(), message))

        if not print_only and self.__base_name and self.__file_dir:
            if action.lower() == 'debug':
                logging.debug(message)
            elif action.lower() == 'info':
                logging.info(message)
            elif action.lower() == 'warning':
                logging.warning(message)
            elif action.lower() == 'error':
                logging.error(message)
            elif action.lower() == 'critical':
                logging.critical(message)

    def __del__(self):
        logging.shutdown()


class GUIConsole(Toplevel):
    __print_lock = Lock()
    __print_queue = list()
    __console_text = None

    """
        Class to show logging in a GUI window by using Toplevel in tkinter
    """

    def __init__(self, parent, title):
        Toplevel.__init__(self)
        self.__parent = parent
        self.bind('<Destroy>', self.__cleanup)

        if title is None:
            self.title("GUI Log")
        else:
            self.title(title)

        self.__build()
        self.__start_idle()

    def __cleanup(self, event):
        self.__parent.gc = None

    def __start_idle(self):
        self.after(5, self.__on_idle)

    def print_gui(self, msg, sep='\n'):
        with self.__print_lock:
            self.__print_queue.append(str(msg) + sep)

    def __build(self):
        my_frame = Frame(self)
        my_frame.pack()

        xbar = Scrollbar(my_frame, orient='horizontal')
        ybar = Scrollbar(my_frame, orient='vertical')
        self.__console_text = Text(my_frame, bg="black", fg="white", wrap="none", yscrollcommand=ybar,
                                   xscrollcommand=xbar)
        xbar.config(command=self.__console_text.xview)
        ybar.config(command=self.__console_text.yview)
        self.__console_text.grid(row=0, column=0)
        xbar.grid(row=1, column=0, sticky=W + E)
        ybar.grid(row=0, column=1, sticky=N + S)

    def __on_idle(self):
        with self.__print_lock:
            for msg in self.__print_queue:
                self.__console_text.insert(END, msg)
                self.__console_text.see(END)

            self.__print_queue = []

        self.after(5, self.__on_idle)



from __future__ import unicode_literals

from tkinter import filedialog, ttk
from tkinter.messagebox import showerror
from tkinter import *
from pickle import dumps

import os


class KSG(Toplevel):
    __pointer = None

    def __init__(self, class_obj, local_dir):
        Toplevel.__init__(self)

        self.__class_obj = class_obj
        self.__local_dir = local_dir
        self.__salt_key_path = StringVar()
        self.__pepper_key_path = StringVar()
        self.__salt_key_dir = StringVar()
        self.__pepper_key_dir = StringVar()

        self.__build()

    def __build(self):
        header = ['Welcome to Salt & Pepper setup!', 'Please fill out the below fields to setup']
        header2 = ['Welcome to existing Salt & Pepper setup!', 'Please fill out the below fields to setup']

        # Set GUI Geometry and GUI Title
        self.geometry('280x215+550+150')
        self.title('Salt & Pepper Setup')
        self.resizable(False, False)

        # Set GUI Frames
        t = ttk.Notebook(self)
        new_tab = ttk.Frame(t)
        existing_tab = ttk.Frame(t)
        t.add(new_tab, text='New Setup')
        t.add(existing_tab, text='Existing')

        new_header_frame = Frame(new_tab)
        new_salt_frame = LabelFrame(new_tab, text='Salt Key', width=508, height=70)
        new_pepper_frame = LabelFrame(new_tab, text='Pepper Key', width=508, height=70)
        new_buttons_frame = Frame(new_tab)
        existing_header_frame = Frame(existing_tab)
        existing_salt_frame = LabelFrame(existing_tab, text='Salt Key', width=508, height=70)
        existing_pepper_frame = LabelFrame(existing_tab, text='Pepper Key', width=508, height=70)
        existing_buttons_frame = Frame(existing_tab)

        # Apply Frames into GUI
        t.pack(fill="both")
        new_header_frame.pack(fill="both")
        new_salt_frame.pack(fill="both")
        new_pepper_frame.pack(fill="both")
        new_buttons_frame.pack(fill="both")
        existing_header_frame.pack(fill="both")
        existing_salt_frame.pack(fill="both")
        existing_pepper_frame.pack(fill="both")
        existing_buttons_frame.pack(fill="both")

        # Apply Header text to Header_Frame that describes purpose of GUI
        header = Message(self, text='\n'.join(header), width=500, justify=CENTER)
        header.pack(in_=new_header_frame)
        header2 = Message(self, text='\n'.join(header2), width=500, justify=CENTER)
        header2.pack(in_=existing_header_frame)

        # Apply Widgets to new_salt_frame
        #   Apply Salt Directory Entry Box Widget
        salt_dir_label = Label(new_salt_frame, text='Salt Dir:')
        salt_dir_entry = Entry(new_salt_frame, width=23, textvariable=self.__salt_key_dir)
        salt_dir_label.grid(row=0, column=0, padx=4, pady=5)
        salt_dir_entry.grid(row=0, column=1, padx=20, pady=5)

        #   Apply Salt Directory Finder Button Widget
        find_salt_dir_button = Button(new_salt_frame, text='...', width=2, command=self.__salt_dir_gui)
        find_salt_dir_button.grid(row=0, column=2, padx=0, pady=5)

        # Apply Widgets to new_settings_frame
        #    Apply Pepper Directory Entry Box Widget
        pepper_dir_label = Label(new_pepper_frame, text='Pepper Dir:')
        pepper_dir_entry = Entry(new_pepper_frame, width=23, textvariable=self.__pepper_key_dir)
        pepper_dir_label.grid(row=0, column=0, padx=4, pady=5)
        pepper_dir_entry.grid(row=0, column=1, padx=4, pady=5)

        #   Apply Pepper Directory Finder Button Widget
        find_pepper_dir_button = Button(new_pepper_frame, text='...', width=2, command=self.__pepper_dir_gui)
        find_pepper_dir_button.grid(row=0, column=2, padx=15, pady=5)

        # Apply Buttons to the new_buttons_frame
        #     Generate button
        generate_button = Button(self, text='Generate', width=15, command=self.__generate)
        generate_button.pack(in_=new_buttons_frame, side=LEFT, padx=7, pady=7)

        #     Cancel button
        cancel_button = Button(self, text='Cancel', width=15, command=self.destroy)
        cancel_button.pack(in_=new_buttons_frame, side=RIGHT, padx=7, pady=7)

        # Apply Widgets to existing_frame
        #   Apply Salt Directory Entry Box Widget
        salt_fp_label = Label(existing_salt_frame, text='Salt FP:')
        salt_fp_entry = Entry(existing_salt_frame, width=23, textvariable=self.__salt_key_path)
        salt_fp_label.grid(row=0, column=0, padx=4, pady=5)
        salt_fp_entry.grid(row=0, column=1, padx=20, pady=5)

        #   Apply Salt File Name Finder Button Widget
        find_salt_name_button = Button(existing_salt_frame, text='...', width=2, command=self.__salt_file_gui)
        find_salt_name_button.grid(row=0, column=2, padx=0, pady=5)

        #   Apply Main Filepath Entry Box Widget
        pepper_fp_label = Label(existing_pepper_frame, text='Pepper FP:')
        pepper_fp_entry = Entry(existing_pepper_frame, width=23, textvariable=self.__pepper_key_path)
        pepper_fp_label.grid(row=1, column=0, padx=4, pady=5)
        pepper_fp_entry.grid(row=1, column=1, padx=4, pady=5)

        #   Apply Main Filepath Finder Button Widget
        find_pepper_fp_button = Button(existing_pepper_frame, text='...', width=2, command=self.__pepper_file_gui)
        find_pepper_fp_button.grid(row=1, column=2, padx=15, pady=5)

        # Apply Buttons to the existing_buttons_frame
        #     Generate button
        link_file_button = Button(self, text='Link File', width=15, command=self.__link_file)
        link_file_button.pack(in_=existing_buttons_frame, side=LEFT, padx=7, pady=7)

        #     Cancel button
        cancel_button = Button(self, text='Cancel', width=15, command=self.destroy)
        cancel_button.pack(in_=existing_buttons_frame, side=RIGHT, padx=7, pady=7)

    def __salt_dir_gui(self):
        if self.__salt_key_dir.get() and os.path.exists(self.__salt_key_dir.get()):
            init_dir = self.__salt_key_dir.get()
        else:
            init_dir = self.__local_dir

        salt_dir = filedialog.askdirectory(initialdir=init_dir, title='Select Salt Key Directory', parent=self)

        if salt_dir:
            self.__salt_key_dir.set(salt_dir)

    def __salt_file_gui(self):
        if self.__salt_key_path.get() and os.path.exists(self.__salt_key_path.get()):
            init_dir = os.path.dirname(self.__salt_key_path.get())
            init_file = os.path.basename(self.__salt_key_path.get())
        else:
            init_dir = self.__local_dir
            init_file = ''

        file_path = filedialog.askopenfile(initialdir=init_dir, initialfile=init_file, title='Select Salt Key',
                                           filetypes=[("key files", "*.key")], parent=self)

        if file_path:
            self.__salt_key_path.set(file_path.name)
            file_path.close()

    def __pepper_dir_gui(self):
        if self.__pepper_key_dir.get() and os.path.exists(self.__pepper_key_dir.get()):
            init_dir = self.__pepper_key_dir.get()
        else:
            init_dir = self.__local_dir

        settings_dir = filedialog.askdirectory(initialdir=init_dir, title='Select Pepper Key Directory',
                                               parent=self)

        if settings_dir:
            self.__pepper_key_dir.set(settings_dir)

    def __pepper_file_gui(self):
        if self.__pepper_key_path.get() and os.path.exists(self.__pepper_key_path.get()):
            init_dir = os.path.dirname(self.__pepper_key_path.get())
            init_file = os.path.basename(self.__pepper_key_path.get())
        else:
            init_dir = self.__local_dir
            init_file = ''

        settings_name = filedialog.askopenfile(initialdir=init_dir, initialfile=init_file,
                                               title='Select Pepper Key', filetypes=[("key files", "*.key")],
                                               parent=self)

        if settings_name:
            self.__pepper_key_path.set(settings_name.name)
            settings_name.close()

    def __link_file(self):
        from . import default_key_dir
        key_dir = default_key_dir()

        if not self.__salt_key_path:
            showerror('Field Empty Error!', 'No value has been inputed for Salt Key Path', parent=self)
        elif not self.__pepper_key_path:
            showerror('Field Empty Error!', 'No value has been inputed for Pepper Key Path', parent=self)
        elif not os.path.isfile(self.__salt_key_path.get()):
            showerror('Invalid Path Error!', 'Salt Key Path is not a file!', parent=self)
        elif not os.path.isfile(self.__pepper_key_path.get()):
            showerror('Invalid Path Error!', 'Pepper Key Path is not a file!', parent=self)
        elif not os.path.exists(self.__salt_key_path.get()):
            showerror('Path Exist Error!', 'Salt Key Path does not exist!', parent=self)
        elif not os.path.exists(self.__pepper_key_path.get()):
            showerror('Path Exist Error!', 'Pepper Key Path does not exist!', parent=self)
        elif os.path.dirname(self.__salt_key_path.get()).lower() == key_dir.lower():
            showerror('Default Path Error!', 'No need to point to default path for Salt Key!', parent=self)
        elif os.path.dirname(self.__pepper_key_path.get()).lower() == key_dir.lower():
            showerror('Default Path Error!', 'No need to point to default path for Pepper Key!', parent=self)
        else:
            from .data import file_write_bytes, KeyPtr

            if not os.path.exists(key_dir):
                os.mkdir(key_dir)

            key_ptr_fp = os.path.join(key_dir, "Key.dir")
            key_ptr = KeyPtr(salt_key_fp=self.__salt_key_path.get(), pepper_key_fp=self.__pepper_key_path.get())
            file_write_bytes(file_path=key_ptr_fp, data=dumps(key_ptr))
            self.__class_obj.key_generated = True

            self.__class_obj.destroy()

    def __generate(self):
        if not self.__salt_key_dir:
            showerror('Field Empty Error!', 'No value has been inputed for Salt Key Directory', parent=self)
        elif not self.__pepper_key_dir:
            showerror('Field Empty Error!', 'No value has been inputed for Pepper Key Directory', parent=self)
        elif not os.path.isdir(self.__salt_key_dir.get()):
            showerror('Invalid Directory!', 'Salt Key Directory is an invalid directory', parent=self)
        elif not os.path.isdir(self.__pepper_key_dir.get()):
            showerror('Invalid Directory!', 'Pepper Key Directory is an invalid directory', parent=self)
        elif not os.path.exists(self.__salt_key_dir.get()):
            showerror('Directory Not Exist!', 'Salt Key Directory does not exist', parent=self)
        elif not os.path.exists(self.__pepper_key_dir.get()):
            showerror('Directory Not Exist!', 'Pepper Key Directory does not exist', parent=self)
        else:
            from .data import create_key, file_write_bytes, KeyPtr
            from . import default_key_dir

            key_dir = default_key_dir()

            if not os.path.exists(key_dir):
                os.mkdir(key_dir)

            key_ptr_fp = os.path.join(key_dir, "Key.dir")
            salt_key_fp = os.path.join(self.__salt_key_dir.get(), "Salt.key")
            pepper_key_fp = os.path.join(self.__pepper_key_dir.get(), "Pepper.key")
            create_key(self.__salt_key_dir.get(), "Salt.key")
            create_key(self.__pepper_key_dir.get(), "Pepper.key")
            key_ptr = KeyPtr(salt_key_fp=salt_key_fp, pepper_key_fp=pepper_key_fp)
            file_write_bytes(key_ptr_fp, dumps(key_ptr))
            self.__class_obj.key_generated = True

            self.__class_obj.destroy()


class KeySetupGUI(KSG):
    """
    Setup GUI for Salt Key, Main DB, Local DB, and Pointer files
    """

    def __init__(self, class_obj, local_dir=None):
        KSG.__init__(self, class_obj, local_dir=local_dir)


class DKSG(Tk):
    __pointer = None

    def __init__(self, local_dir):
        Tk.__init__(self)

        self.__local_dir = local_dir
        self.__key_generated = False
        self.__specific_obj = None

        self.__build()

    @property
    def key_generated(self):
        return self.__key_generated

    @key_generated.setter
    def key_generated(self, key_generated):
        self.__key_generated = key_generated

    def __build(self):
        header = ['Welcome to Salt & Pepper setup!', 'Would you like to install in default location?']

        # Set GUI Geometry and GUI Title
        self.geometry('280x80+550+150')
        self.title('Salt & Pepper Setup')
        self.resizable(False, False)

        new_header_frame = Frame(self)
        new_buttons_frame = Frame(self)

        # Apply Frames into GUI
        new_header_frame.pack(fill="both")
        new_buttons_frame.pack(fill="both")

        # Apply Header text to Header_Frame that describes purpose of GUI
        header = Message(self, text='\n'.join(header), width=500, justify=CENTER)
        header.pack(in_=new_header_frame)

        # Apply Buttons to the new_buttons_frame
        #     Default button
        default_button = Button(self, text='Default Location', width=15, command=self.__default)
        default_button.pack(in_=new_buttons_frame, side=LEFT, padx=7, pady=7)

        #     Cancel button
        specific_button = Button(self, text='Specific Location', width=15, command=self.__specific)
        specific_button.pack(in_=new_buttons_frame, side=RIGHT, padx=7, pady=7)

    def __default(self):
        if not self.__specific_obj:
            from . import default_key_dir
            from .data import create_key
            key_dir = default_key_dir()

            create_key(key_dir, "Salt.key")
            create_key(key_dir, "Pepper.key")
            self.__key_generated = True
            self.destroy()

    def __specific(self):
        self.__specific_obj = KeySetupGUI(class_obj=self, local_dir=self.__local_dir)


class DefaultKeySetupGUI(DKSG):
    """
    Default Setup GUI for Salt Key, Main DB, Local DB, and Pointer files
    """

    def __init__(self, local_dir=None):
        DKSG.__init__(self, local_dir=local_dir)
        self.mainloop()


class MDG(Tk):
    __pointer = None

    def __init__(self, local_db_dir, main_db_path=None, pointer=None):
        Tk.__init__(self)

        from .data import DataConfig

        if not isinstance(pointer, (DataConfig, type(None))):
            raise ValueError("'pointer' %r is not an instance of DataConfig" % pointer)
        if not isinstance(local_db_dir, str):
            raise ValueError("'local_db_dir' %r is not an instance of String" % local_db_dir)
        if not isinstance(main_db_path, (str, type(None))):
            raise ValueError("'main_db_path' %r is not an instance of String" % main_db_path)
        if not os.path.isdir(local_db_dir):
            raise ValueError("'local_db_dir' %r is not a valid directory" % local_db_dir)
        if main_db_path and not os.path.exists(main_db_path):
            main_db_path = os.path.dirname(main_db_path)

        self.__pointer = pointer
        self.__local_dir = local_db_dir
        self.__main_db_path = StringVar()
        self.__main_db_dir = StringVar()
        self.__local_db_path = StringVar()
        self.__local_db_dir = StringVar()

        if main_db_path:
            self.__main_db_path.set(main_db_path)

        self.__build()

    @property
    def pointer(self):
        """
        :return: Returns Pointer DataConfig object
        """

        return self.__pointer

    def __build(self):
        header = ['Welcome to new database setup!', 'Please fill out the below fields to setup']
        header2 = ['Welcome to existing database setup!', 'Please fill out the below fields to setup']

        # Set GUI Geometry and GUI Title
        self.geometry('280x195+550+150')
        self.title('Main Database Setup')
        self.resizable(False, False)

        # Set GUI Frames
        t = ttk.Notebook(self)
        new_tab = ttk.Frame(t)
        existing_tab = ttk.Frame(t)
        t.add(new_tab, text='New Setup')
        t.add(existing_tab, text='Existing')

        new_header_frame = Frame(new_tab)
        new_settings_frame = LabelFrame(new_tab, text='Create DBs', width=508, height=70)
        new_buttons_frame = Frame(new_tab)
        existing_header_frame = Frame(existing_tab)
        existing_main_frame = LabelFrame(existing_tab, text='Find DBs', width=508, height=70)
        existing_buttons_frame = Frame(existing_tab)

        # Apply Frames into GUI
        t.pack(fill="both")
        new_header_frame.pack(fill="both")
        new_settings_frame.pack(fill="both")
        new_buttons_frame.pack(fill="both")
        existing_header_frame.pack(fill="both")
        existing_main_frame.pack(fill="both")
        existing_buttons_frame.pack(fill="both")

        # Apply Header text to Header_Frame that describes purpose of GUI
        header = Message(self, text='\n'.join(header), width=500, justify=CENTER)
        header.pack(in_=new_header_frame)
        header2 = Message(self, text='\n'.join(header2), width=500, justify=CENTER)
        header2.pack(in_=existing_header_frame)

        # Apply Widgets to new_settings_frame
        #    Apply Settings Directory Entry Box Widget
        settings_dir_label = Label(new_settings_frame, text='Main DB Dir:')
        settings_dir_entry = Entry(new_settings_frame, textvariable=self.__main_db_dir, width=23)
        settings_dir_label.grid(row=0, column=0, padx=4, pady=5)
        settings_dir_entry.grid(row=0, column=1, padx=4, pady=5)

        #   Apply Settings Directory Finder Button Widget
        find_settings_dir_button = Button(new_settings_frame, text='...', width=3, command=self.__main_dir_gui)
        find_settings_dir_button.grid(row=0, column=2, padx=4, pady=5)

        #    Apply Settings Directory Entry Box Widget
        settings_dir_label = Label(new_settings_frame, text='Local DB Dir:')
        settings_dir_entry = Entry(new_settings_frame, textvariable=self.__local_db_dir, width=23)
        settings_dir_label.grid(row=1, column=0, padx=4, pady=5)
        settings_dir_entry.grid(row=1, column=1, padx=4, pady=5)

        #   Apply Settings Directory Finder Button Widget
        find_settings_dir_button = Button(new_settings_frame, text='...', width=3, command=self.__local_dir_gui)
        find_settings_dir_button.grid(row=1, column=2, padx=4, pady=5)

        # Apply Buttons to the new_buttons_frame
        #     Generate button
        generate_button = Button(self, text='Generate', width=15, command=self.__generate)
        generate_button.pack(in_=new_buttons_frame, side=LEFT, padx=7, pady=7)

        #     Cancel button
        cancel_button = Button(self, text='Cancel', width=15, command=self.__destroy)
        cancel_button.pack(in_=new_buttons_frame, side=RIGHT, padx=7, pady=7)

        # Apply Widgets to existing_frame
        #   Apply Main Filepath Entry Box Widget
        settings_fp_label = Label(existing_main_frame, text='Main DB Fp:')
        settings_fp_entry = Entry(existing_main_frame, textvariable=self.__main_db_path, width=23)
        settings_fp_label.grid(row=0, column=0, padx=4, pady=5)
        settings_fp_entry.grid(row=0, column=1, padx=4, pady=5)

        #   Apply Main Filepath Finder Button Widget
        find_settings_name_button = Button(existing_main_frame, text='...', width=3, command=self.__main_file_gui)
        find_settings_name_button.grid(row=0, column=2, padx=4, pady=5)

        #   Apply Main Filepath Entry Box Widget
        settings_fp_label = Label(existing_main_frame, text='Local DB Fp:')
        settings_fp_entry = Entry(existing_main_frame, textvariable=self.__local_db_path, width=23)
        settings_fp_label.grid(row=1, column=0, padx=4, pady=5)
        settings_fp_entry.grid(row=1, column=1, padx=4, pady=5)

        #   Apply Main Filepath Finder Button Widget
        find_settings_name_button = Button(existing_main_frame, text='...', width=3, command=self.__local_file_gui)
        find_settings_name_button.grid(row=1, column=2, padx=4, pady=5)

        # Apply Buttons to the existing_buttons_frame
        #     Generate button
        generate_button = Button(self, text='Link DBs', width=15, command=self.__generate2)
        generate_button.pack(in_=existing_buttons_frame, side=LEFT, padx=7, pady=7)

        #     Cancel button
        cancel_button = Button(self, text='Cancel', width=15, command=self.__destroy)
        cancel_button.pack(in_=existing_buttons_frame, side=RIGHT, padx=7, pady=7)

    def __destroy(self):
        self.__pointer = None
        self.destroy()

    def __main_dir_gui(self):
        if self.__main_db_dir.get() and os.path.exists(self.__main_db_dir.get()):
            init_dir = self.__main_db_dir.get()
        else:
            init_dir = self.__local_dir

        settings_dir = filedialog.askdirectory(initialdir=init_dir, title='Select Main DB Directory', parent=self)

        if settings_dir:
            self.__main_db_dir.set(settings_dir)

    def __local_dir_gui(self):
        if self.__local_db_dir.get() and os.path.exists(self.__local_db_dir.get()):
            init_dir = self.__local_db_dir.get()
        else:
            init_dir = self.__local_dir

        settings_dir = filedialog.askdirectory(initialdir=init_dir, title='Select Local DB Directory', parent=self)

        if settings_dir:
            self.__local_db_dir.set(settings_dir)

    def __main_file_gui(self):
        if self.__main_db_path.get() and os.path.exists(self.__main_db_path.get()):
            init_dir = os.path.dirname(self.__main_db_path.get())
            init_file = os.path.basename(self.__main_db_path.get())
        else:
            init_dir = self.__local_dir
            init_file = ''

        settings_name = filedialog.askopenfile(initialdir=init_dir, initialfile=init_file,
                                               title='Select Main DB', filetypes=[("db files", "*.db")],
                                               parent=self)

        if settings_name:
            self.__main_db_path.set(settings_name.name)
            settings_name.close()

    def __local_file_gui(self):
        if self.__local_db_path.get() and os.path.exists(self.__local_db_path.get()):
            init_dir = os.path.dirname(self.__local_db_path.get())
            init_file = os.path.basename(self.__local_db_path.get())
        else:
            init_dir = self.__local_dir
            init_file = ''

        settings_name = filedialog.askopenfile(initialdir=init_dir, initialfile=init_file,
                                               title='Select Local DB', filetypes=[("db files", "*.db")],
                                               parent=self)

        if settings_name:
            self.__local_db_path.set(settings_name.name)
            settings_name.close()

    def __generate2(self):
        if not self.__main_db_path:
            showerror('Field Empty Error!', 'No value has been inputed for Main DB Path', parent=self)
        elif not os.path.isfile(self.__main_db_path.get()) and not os.path.isdir(self.__main_db_path.get()):
            showerror('Invalid Path Error!', 'Main DB Path is not a file or directory!', parent=self)
        elif not os.path.exists(self.__main_db_path.get()):
            showerror('Path Exist Error!', 'Main DB Path does not exist!', parent=self)
        elif not self.__local_db_path:
            showerror('Field Empty Error!', 'No value has been inputed for Local DB Path', parent=self)
        elif not os.path.isfile(self.__local_db_path.get()) and not os.path.isdir(self.__local_db_path.get()):
            showerror('Invalid Path Error!', 'Local DB Path is not a file or directory!', parent=self)
        elif not os.path.exists(self.__local_db_path.get()):
            showerror('Path Exist Error!', 'Local DB Path does not exist!', parent=self)
        else:
            if not self.__pointer:
                from .data import DataConfig

                self.__pointer = DataConfig(file_dir=self.__local_dir, file_name_prefix='Script_Pointer',
                                            file_ext='ptr', encrypt=True)

            if os.path.isfile(self.__local_db_path.get()):
                self.__pointer.setcrypt(key='Local_DB_Path', val=self.__local_db_path.get(), private=True)
            else:
                self.__pointer.setcrypt(key='Local_DB_Path', val=os.path.join(self.__local_db_path.get(),
                                                                             'Local_Settings.db'), private=True)

            if os.path.isfile(self.__main_db_path.get()):
                self.__pointer.setcrypt(key='Main_DB_Path', val=self.__main_db_path.get(), private=True)
            else:
                self.__pointer.setcrypt(key='Main_DB_Path', val=os.path.join(self.__main_db_path.get(),
                                                                             'Main_Settings.db'), private=True)
            self.__pointer.sync()
            self.destroy()

    def __generate(self):
        if not self.__main_db_dir:
            showerror('Field Empty Error!', 'No value has been inputed for Main DB Directory', parent=self)
        elif not os.path.isdir(self.__main_db_dir.get()):
            showerror('Invalid Directory!', 'Main DB Directory is an invalid directory', parent=self)
        elif not os.path.exists(self.__main_db_dir.get()):
            showerror('Directory Not Exist!', 'Main DB Directory does not exist', parent=self)
        elif not self.__local_db_dir:
            showerror('Field Empty Error!', 'No value has been inputed for Local DB Directory', parent=self)
        elif not os.path.isdir(self.__local_db_dir.get()):
            showerror('Invalid Directory!', 'Local DB Directory is an invalid directory', parent=self)
        elif not os.path.exists(self.__local_db_dir.get()):
            showerror('Directory Not Exist!', 'Local DB Directory does not exist', parent=self)
        else:
            if not self.__pointer:
                from .data import DataConfig

                self.__pointer = DataConfig(file_dir=self.__local_dir, file_name_prefix='Script_Pointer',
                                            file_ext='ptr', encrypt=True)

            self.__pointer.setcrypt(key='Local_DB_Path', val=os.path.join(self.__local_db_dir.get(),
                                                                          'Local_Settings.db'), private=True)
            self.__pointer.setcrypt(key='Main_DB_Path', val=os.path.join(self.__main_db_dir.get(), 'Main_Settings.db'),
                                    private=True)
            self.__pointer.sync()
            self.destroy()


class MainDatabaseGUI(MDG):
    """
    Setup GUI for Salt Key, Main DB, Local DB, and Pointer files
    """

    def __init__(self, local_db_dir, main_db_path=None, pointer=None):
        """
        :param local_db_dir:  Local DB directory (aka local script directory)
        :param main_db_path: (Optional) Main DB filepath
        :param pointer: (Optional) Pointer .ptr DataConfig instance class
        """

        MDG.__init__(self, local_db_dir=local_db_dir, main_db_path=main_db_path, pointer=pointer)
        self.mainloop()


class CG(Tk):
    def __init__(self, cred=None, user_name=None, user_pass=None):
        Tk.__init__(self)

        from .credentials import Credentials
        from .data import CryptHandle

        if not isinstance(cred, (Credentials, type(None))):
            raise ValueError("'cred' %r is not an instance of Credentials" % cred)
        if not isinstance(user_name, (str, type(None))):
            raise ValueError("'user_name' %r is not a String" % user_name)
        if not isinstance(user_pass, (str, type(None))):
            raise ValueError("'user_pass' %r is not a String" % user_pass)

        self.__cred = cred
        self.__user_name = StringVar()
        self.__user_pass = StringVar()

        if cred:
            self.__user_name_obj = self.__cred.username
            self.__user_pass_obj = self.__cred.password
            self.__user_name.set(self.__user_name_obj.peak())
            self.__user_pass.set(self.__user_pass_obj.peak())
        else:
            self.__user_name_obj = CryptHandle(private=False)
            self.__user_pass_obj = CryptHandle(private=True)

            if user_name:
                self.__user_name_obj.encrypt(user_name)
                self.__user_name.set(self.__user_name_obj.peak())

            if user_pass:
                self.__user_pass_obj.encrypt(user_pass)
                self.__user_pass.set(self.__user_pass_obj.peak())

        self.__build()

    @property
    def credentials(self):
        """
        :return: Returns Credentials instance class, which holds username & password in CryptHandle objects
        """

        return self.__cred

    def __build(self):
        header = ['Welcome to Credential Creating', 'Please fill out the fields below']

        # Set GUI Geometry and GUI Title
        self.geometry('217x160+650+280')
        self.title('Credentials Setup')
        self.resizable(False, False)

        # Set GUI Frames
        header_frame = Frame(self)
        cred_frame = LabelFrame(self, text='Credentials', width=508, height=70)
        buttons_frame = Frame(self)

        # Apply Frames into GUI
        header_frame.pack(fill="both")
        cred_frame.pack(fill="both")
        buttons_frame.pack(fill="both")

        # Apply Header text to Header_Frame that describes purpose of GUI
        header = Message(self, text='\n'.join(header), width=500, justify=CENTER)
        header.pack(in_=header_frame)

        # Apply Widgets to Cred Frame
        #   Apply User Name Entry Box Widget
        user_name_label = Label(cred_frame, text='User Name:')
        user_name_entry = Entry(cred_frame, textvariable=self.__user_name)
        user_name_label.grid(row=0, column=0, padx=4, pady=5)
        user_name_entry.grid(row=0, column=1, padx=4, pady=5)

        #   Apply User Pass Entry Box Widget
        user_pass_label = Label(cred_frame, text='User Pass:')
        user_pass_entry = Entry(cred_frame, textvariable=self.__user_pass)
        user_pass_label.grid(row=1, column=0, padx=4, pady=5)
        user_pass_entry.grid(row=1, column=1, padx=4, pady=5)
        user_pass_entry.bind('<KeyRelease>', self.__hide_pass)

        # Apply Buttons to the Buttons Frame
        #     Save button
        save_button = Button(self, text='Save', width=13, command=self.__save)
        save_button.pack(in_=buttons_frame, side=LEFT, padx=7, pady=7)

        #     Cancel button
        cancel_button = Button(self, text='Cancel', width=13, command=self.destroy)
        cancel_button.pack(in_=buttons_frame, side=RIGHT, padx=7, pady=7)

    def __hide_pass(self, event):
        new_pass = self.__user_pass.get()
        curr_pass = self.__user_pass_obj.decrypt()

        if not curr_pass:
            curr_pass = ''

        curr_pass_len = len(curr_pass)

        if len(new_pass) > 0:
            curr_pass = self.__adjust_pass(new_pass, curr_pass)

            if curr_pass_len > len(new_pass):
                curr_pass = curr_pass[:len(new_pass)]
        else:
            curr_pass = new_pass

        self.__user_pass_obj.encrypt(curr_pass)
        self.__user_pass.set(self.__user_pass_obj.peak())

    @staticmethod
    def __adjust_pass(new_pass, curr_pass):
        curr_pass_len = len(curr_pass)

        for pos, letter in enumerate(new_pass):
            if letter != '*':
                if pos > curr_pass_len - 1:
                    curr_pass += letter
                else:
                    my_text = list(curr_pass)
                    my_text.insert(pos, letter)
                    curr_pass = ''.join(my_text)

        return curr_pass

    def __save(self):
        if not self.__user_name:
            showerror('Field Empty Error!', 'No value has been inputed for User Name', parent=self)
        elif not self.__user_pass:
            showerror('Field Empty Error!', 'No value has been inputed for User Pass', parent=self)
        else:
            from . import Credentials

            self.__user_name_obj.encrypt(self.__user_name.get())
            self.__cred = Credentials(user_name=self.__user_name_obj, user_pass=self.__user_pass_obj)
            self.destroy()


class CredentialsGUI(CG):
    """
    Credentials GUI for gathering Username and Password, and storing this information in encrypted objects within
    the Credentials instance class
    """

    def __init__(self, cred=None, user_name=None, user_pass=None):
        """

        :param cred: (Optional) Credentials instance class object
        :param user_name: (Optional) Username in String
        :param user_pass: (Optional) Userpass in String
        """

        CG.__init__(self, cred=cred, user_name=user_name, user_pass=user_pass)
        self.mainloop()


class EG(Tk):
    def __init__(self, server=None, email_addr=None):
        Tk.__init__(self)

        if not isinstance(server, (str, type(None))):
            raise ValueError("'server' %r is not a String" % server)
        if not isinstance(email_addr, (str, type(None))):
            raise ValueError("'email_addr' %r is not a String" % email_addr)

        from .data import CryptHandle

        self.__server_obj = CryptHandle()
        self.__email_addr_obj = CryptHandle()
        self.__server = StringVar()
        self.__email_addr = StringVar()

        if server:
            self.__server.set(server)

        if email_addr:
            self.__email_addr.set(email_addr)

        self.__build()

    @property
    def server(self):
        return self.__server_obj

    @property
    def email_addr(self):
        return self.__email_addr_obj

    def __build(self):
        header = ['Welcome to Email Settings!', 'Please fill out the fields below:']

        # Set GUI Geometry and GUI Title
        self.geometry('217x160+650+280')
        self.title('Credentials Setup')
        self.resizable(False, False)

        # Set GUI Frames
        header_frame = Frame(self)
        esettings_frame = LabelFrame(self, text='Email Settings', width=508, height=70)
        buttons_frame = Frame(self)

        # Apply Frames into GUI
        header_frame.pack(fill="both")
        esettings_frame.pack(fill="both")
        buttons_frame.pack(fill="both")

        # Apply Header text to Header_Frame that describes purpose of GUI
        header = Message(self, text='\n'.join(header), width=500, justify=CENTER)
        header.pack(in_=header_frame)

        # Apply Widgets to Cred Frame
        #   Apply User Name Entry Box Widget
        server_label = Label(esettings_frame, text='Server:')
        server_entry = Entry(esettings_frame, textvariable=self.__server)
        server_label.grid(row=0, column=0, padx=4, pady=5)
        server_entry.grid(row=0, column=1, padx=4, pady=5)

        #   Apply User Pass Entry Box Widget
        email_addr_label = Label(esettings_frame, text='Email Addr:')
        email_addr_entry = Entry(esettings_frame, textvariable=self.__email_addr)
        email_addr_label.grid(row=1, column=0, padx=4, pady=5)
        email_addr_entry.grid(row=1, column=1, padx=4, pady=5)

        # Apply Buttons to the Buttons Frame
        #     Save button
        save_button = Button(self, text='Save', width=13, command=self.__save)
        save_button.pack(in_=buttons_frame, side=LEFT, padx=7, pady=7)

        #     Cancel button
        cancel_button = Button(self, text='Cancel', width=13, command=self.destroy)
        cancel_button.pack(in_=buttons_frame, side=RIGHT, padx=7, pady=7)

    def __save(self):
        if not self.__server:
            showerror('Field Empty Error!', 'No value has been inputed for Server', parent=self)
        elif not self.__email_addr:
            showerror('Field Empty Error!', 'No value has been inputed for Email Addr', parent=self)
        elif not self.__email_addr.get().find('@'):
            showerror('Field Bad Result Error!', 'Email Addr does not have an @ in it')
        else:
            self.__server_obj.encrypt(self.__server.get())
            self.__email_addr_obj.encrypt(self.__email_addr.get())
            self.destroy()


class EmailGUI(EG):
    """
    Email GUI to find Server Address and Email address for email exchange connections
    """

    def __init__(self, server=None, email_addr=None):
        """

        :param server: (Optional) Server address in string format
        :param email_addr: (Optional) Email address in string format
        """
        EG.__init__(self, server=server, email_addr=email_addr)
        self.mainloop()


class SSG(Tk):
    def __init__(self, server=None, database=None):
        Tk.__init__(self)
        from .data import CryptHandle

        if not isinstance(server, (str, type(None))):
            raise ValueError("'server' %r is not a String" % server)
        if not isinstance(database, (str, type(None))):
            raise ValueError("'database' %r is not a String" % database)

        self.__server = StringVar()
        self.__database = StringVar()
        self.__server_obj = CryptHandle()
        self.__database_obj = CryptHandle()

        if server:
            self.__server.set(server)
        if database:
            self.__database.set(database)

        self.__build()

    @property
    def server(self):
        """
        :return: Returns CryptHandle object of server
        """

        return self.__server_obj

    @property
    def database(self):
        """
        :return: Returns CryptHandle object of database
        """

        return self.__database_obj

    def __build(self):
        header = ['Welcome to SQL Server Settings!', 'Please fill out the fields below:']

        # Set GUI Geometry and GUI Title
        self.geometry('217x160+650+280')
        self.title('SQL Server Setup')
        self.resizable(False, False)

        # Set GUI Frames
        header_frame = Frame(self)
        ssettings_frame = LabelFrame(self, text='SQL Settings', width=508, height=70)
        buttons_frame = Frame(self)

        # Apply Frames into GUI
        header_frame.pack(fill="both")
        ssettings_frame.pack(fill="both")
        buttons_frame.pack(fill="both")

        # Apply Header text to Header_Frame that describes purpose of GUI
        header = Message(self, text='\n'.join(header), width=500, justify=CENTER)
        header.pack(in_=header_frame)

        # Apply Widgets to Cred Frame
        #   Apply User Name Entry Box Widget
        server_label = Label(ssettings_frame, text='Server:')
        server_entry = Entry(ssettings_frame, textvariable=self.__server)
        server_label.grid(row=0, column=0, padx=4, pady=5)
        server_entry.grid(row=0, column=1, padx=4, pady=5)

        #   Apply User Pass Entry Box Widget
        database_label = Label(ssettings_frame, text='Database:')
        database_entry = Entry(ssettings_frame, textvariable=self.__database)
        database_label.grid(row=1, column=0, padx=4, pady=5)
        database_entry.grid(row=1, column=1, padx=4, pady=5)

        # Apply Buttons to the Buttons Frame
        #     Save button
        save_button = Button(self, text='Save', width=13, command=self.__save)
        save_button.pack(in_=buttons_frame, side=LEFT, padx=7, pady=7)

        #     Cancel button
        cancel_button = Button(self, text='Cancel', width=13, command=self.destroy)
        cancel_button.pack(in_=buttons_frame, side=RIGHT, padx=7, pady=7)

    def __save(self):
        if not self.__server:
            showerror('Field Empty Error!', 'No value has been inputed for Server', parent=self)
        elif not self.__database:
            showerror('Field Empty Error!', 'No value has been inputed for Database', parent=self)
        else:
            from . import Credentials

            self.__server_obj.encrypt(self.__server.get())
            self.__database_obj.encrypt(self.__database.get())
            self.destroy()


class SQLServerGUI(SSG):
    """
    SQL Server GUI that gathers Server & Database for basic SQL connections
    """

    def __init__(self, server=None, database=None):
        """
        :param server: Server in string format
        :param database: Database in string format
        """

        SSG.__init__(self, server=server, database=database)
        self.mainloop()



from pandas import DataFrame, ExcelFile
from os.path import exists, isdir
from xml.etree.ElementTree import parse as xml_parse
from django.utils.crypto import get_random_string
from pandas.core.dtypes.common import is_list_like
from pandas.io.parsers import TextParser
from pandas.errors import EmptyDataError
from portalocker.utils import DEFAULT_TIMEOUT, DEFAULT_CHECK_INTERVAL, DEFAULT_FAIL_WHEN_LOCKED, LOCK_METHOD
from typing import Mapping, Any


class NotFunctionError(Exception):
    """
        Exception raised for variable not being a callable function
    """
    pass


class Streams(object):
    __engines = ['xmlread', 'xmlwrite', 'csv', 'open', 'xlrd', 'openpyxl', 'odf', 'pyxlsb', 'default']

    __slots__ = "streams"

    def __init__(self):
        self.streams = dict()

    def add_stream(self, engine, handler, buffer):
        assert engine in self.__engines, f"Engine {engine} not recognized"

        if engine not in self.streams.keys():
            self.streams[engine] = list()

        self.streams[engine].append([handler, buffer])

    def keys(self):
        return list(self.streams.keys())

    def items(self):
        return self.streams.items()

    def __getitem__(self, key):
        if isinstance(key, str) and key in self.streams.keys():
            return self.streams[key]
        else:
            return list()

    def __setitem__(self, key, value):
        if isinstance(key, str):
            self.streams[key] = value

    def __delitem__(self, key):
        if isinstance(key, str) and key in self.streams.keys():
            del self.streams[key]

    def __contains__(self, key):
        if isinstance(key, str):
            return key in self.streams.keys()

    def __iter__(self):
        for k in self.streams.keys():
            yield k

    def __len__(self):
        return len(self.streams)

    def __repr__(self):
        return self.__class__.__name__ + repr(str(self.streams))

    def __str__(self):
        return str(self.streams)

    def __eq__(self, other):
        for k in self.__slots__:
            if getattr(self, k) != getattr(other, k):
                return False

        return True


class FileHandler(object):
    __file_path = None
    __file_dir = None
    __sheet_names = list()
    __usecols = None
    __header = 0
    __streams = list()
    __df = DataFrame()

    def __init__(self, file_path=None, file_dir=None, file_name=None, flag_read_only=False, sheet_names=None, header=0,
                 names=None, index_col=None, usecols=None, squeeze=False, dtype=None, true_values=None,
                 false_values=None, skiprows=0, nrows=0, na_values=None, verbose=False, parse_dates=False,
                 date_parser=None, thousands=None, comment=None, skipfooter=0, convert_float=True, dialect=None,
                 mangle_dupe_cols=True, xmlns_rs=None, dict_var=None, df=None, mode='a', timeout=DEFAULT_TIMEOUT,
                 check_interval=DEFAULT_CHECK_INTERVAL, fail_when_locked=DEFAULT_FAIL_WHEN_LOCKED, flags=LOCK_METHOD,
                 **kwargs):
        self.file_path = file_path
        self.file_dir = file_dir
        self.file_name = file_name
        self.flag_read_only = flag_read_only
        self.sheet_names = sheet_names
        self.header = header
        self.names = names
        self.index_col = index_col
        self.usecols = usecols
        self.squeeze = squeeze
        self.dtype = dtype
        self.true_values = true_values
        self.false_values = false_values
        self.skiprows = skiprows
        self.nrows = nrows
        self.na_values = na_values
        self.verbose = verbose
        self.parse_dates = parse_dates
        self.date_parser = date_parser
        self.thousands = thousands
        self.comment = comment
        self.skipfooter = skipfooter
        self.convert_float = convert_float
        self.mangle_dupe_cols = mangle_dupe_cols
        self.xmlns_rs = xmlns_rs
        self.dict_var = dict_var
        self.df = df
        self.mode = mode
        self.timeout = timeout,
        self.check_interval = check_interval
        self.fail_when_locked = fail_when_locked
        self.flags = flags
        self.dialect = dialect
        self.kwargs = kwargs

    @property
    def file_path(self):
        return self.__file_path

    @file_path.setter
    def file_path(self, file_path):
        if file_path:
            if not exists(file_path):
                raise ValueError("'file_path' %r does not exist" % file_path)

            self.__file_path = file_path
        else:
            self.__file_path = None

    @property
    def file_dir(self):
        return self.__file_dir

    @file_dir.setter
    def file_dir(self, file_dir):
        if file_dir:
            if not isdir(file_dir):
                raise ValueError("'file_dir' %r is not a valid directory" % file_dir)
            if not exists(file_dir):
                raise Exception("'xml_file' %r does not exist" % file_dir)

            self.__file_dir = file_dir
        else:
            self.__file_dir = None

    @property
    def sheet_names(self):
        return self.__sheet_names

    @sheet_names.setter
    def sheet_names(self, sheet_names):
        if sheet_names:
            if not isinstance(sheet_names, list):
                raise ValueError("'sheet_names' %r is not a list value" % sheet_name)

            self.__sheet_names = [s.lower() for s in self.sheet_names]
        else:
            self.__sheet_names = list()

    @property
    def usecols(self):
        return self.__usecols

    @usecols.setter
    def usecols(self, usecols):
        if usecols:
            self.__usecols = maybe_convert_usecols(usecols)
        else:
            self.__usecols = None

    @property
    def header(self):
        return self.__header

    @header.setter
    def header(self, header):
        if header:
            if not isinstance(header, int):
                raise ValueError("'header' %r is not a integer" % header)

            self.__header = header
        else:
            self.__header = 0

    @property
    def streams(self):
        return self.__streams

    @streams.setter
    def streams(self, streams):
        if streams:
            if not isinstance(streams, list):
                raise ValueError("'streams' %r is not a list" % streams)

            self.__streams = streams
        else:
            self.__streams = list()

    @property
    def df(self):
        return self.__df

    @df.setter
    def df(self, df):
        if df:
            if not isinstance(df, DataFrame):
                raise ValueError("'df' %r is not an instance of DataFrame" % df)

            self.__df = df
        else:
            self.__df = DataFrame()

    def convert_data(self, data):
        if data:
            if self.nrows == 0:
                nrows = None
            else:
                nrows = self.nrows

            if is_list_like(self.index_col):
                if self.header is None:
                    offset = 0
                else:
                    offset = 1 + self.header

                if offset < len(data):
                    for col in self.index_col:
                        last = data[offset][col]

                        for row in range(offset + 1, len(data)):
                            if data[row][col] == "" or data[row][col] is None:
                                data[row][col] = last
                            else:
                                last = data[row][col]

            try:
                parser = TextParser(
                    data,
                    names=self.names,
                    header=0,
                    index_col=self.index_col,
                    has_index_names=False,
                    squeeze=self.squeeze,
                    dtype=self.dtype,
                    true_values=self.true_values,
                    false_values=self.false_values,
                    skiprows=None,
                    nrows=nrows,
                    na_values=self.na_values,
                    parse_dates=self.parse_dates,
                    date_parser=self.date_parser,
                    thousands=self.thousands,
                    comment=self.comment,
                    skipfooter=self.skipfooter,
                    usecols=self.usecols,
                    mangle_dupe_cols=self.mangle_dupe_cols,
                    **self.kwargs,
                )

                df = parser.read(nrows=nrows)
            except EmptyDataError:
                df = DataFrame()

            return df
        else:
            return DataFrame()


class FileParser(object):
    from .reader._xml import XMLReader, XMLWriter
    from .reader._open import OpenReader
    from .reader._excel import OpenPYXLReader, ODFReader, XLRDReader, PYXLSBReader
    from .reader._csv import CSVReader
    from .reader._json import JSONReader

    DEFAULTBUFFER = 1000
    __streams = Streams()
    __reader = None
    __engines: Mapping[str, Any] = {
        "xmlread": XMLReader,
        "xmlwrite": XMLWriter,
        "csv": CSVReader,
        "open": OpenReader,
        "xlrd": XLRDReader,
        "openpyxl": OpenPYXLReader,
        "odf": ODFReader,
        "pyxlsb": PYXLSBReader,
        "json": JSONReader
    }

    @property
    def reader(self):
        return self.__reader

    @reader.setter
    def reader(self, engine_or_reader):
        if engine_or_reader:
            if isinstance(engine_or_reader, str):
                assert engine_or_reader in self.__engines, f"Engine {engine_or_reader} not recognized"
                engine_or_reader = self.__engines[engine_or_reader]

            self.__reader = engine_or_reader
        else:
            self.__reader = None

    def stream_data(self, buffer=DEFAULTBUFFER, engine='default'):
        def registerhandler(handler):
            self.__streams.add_stream(engine, handler, buffer)
            return handler

        return registerhandler

    def parse_excel(self, file_path, flag_read_only=False, sheet_names=None, header=0, names=None, index_col=None,
                    usecols=None, squeeze=False, converters=None, dtype=None, engine=None, true_values=None,
                    false_values=None, skiprows=0, nrows=0, na_values=None, keep_default_na=True, na_filter=True,
                    verbose=False, parse_dates=False, date_parser=None, thousands=None, comment=None, skipfooter=0,
                    convert_float=True, mangle_dupe_cols=True):

        obj = ExcelFile(file_path, engine)
        self.reader = self.__engines[obj.engine](file_path=file_path, flag_read_only=flag_read_only,
                                                 sheet_names=sheet_names, header=header, names=names,
                                                 index_col=index_col, usecols=usecols, squeeze=squeeze, dtype=dtype,
                                                 true_values=true_values, false_values=false_values, skiprows=skiprows,
                                                 nrows=nrows, na_values=na_values, verbose=verbose,
                                                 parse_dates=parse_dates, date_parser=date_parser, thousands=thousands,
                                                 comment=comment, skipfooter=skipfooter, convert_float=convert_float,
                                                 mangle_dupe_cols=mangle_dupe_cols, converters=converters,
                                                 keep_default_na=keep_default_na, na_filter=na_filter)
        self.reader.streams = self.__streams[obj.engine] + self.__streams['default']
        return self.reader.parse()

    def parse_xml(self, file_path, xmlns_rs, dict_var=None):
        self.reader = self.__engines['xmlread'](file_path=file_path, xmlns_rs=xmlns_rs, dict_var=dict_var)
        self.reader.streams = self.__streams['xmlread'] + self.__streams['default']
        return self.reader.parse()

    def write_xml(self, file_dir, file_name, df):
        self.reader = self.__engines['xmlwriter'](file_dir=file_dir, file_name=file_name, df=df)
        self.reader.write()

    def parse_open(self, file_path, mode='a', timeout=DEFAULT_TIMEOUT, check_interval=DEFAULT_CHECK_INTERVAL,
                   fail_when_locked=DEFAULT_FAIL_WHEN_LOCKED, flags=LOCK_METHOD, **file_open_kwargs):
        self.reader = self.__engines['open'](file_path=file_path, mode=mode, timeout=timeout,
                                             check_interval=check_interval, fail_when_locked=fail_when_locked,
                                             flags=flags, **file_open_kwargs)
        self.reader.streams = self.__streams['open'] + self.__streams['default']
        return self.reader.parse()

    def parse_csv(self, file_path, dialect=None, **fmtparams):
        self.reader = self.__engines['csv'](file_path=file_path, dialect=dialect, **fmtparams)
        self.reader.streams = self.__streams['csv'] + self.__streams['default']
        return self.reader.parse()

    def parse_json(self, file_path):
        self.reader = self.__engines['json'](file_path=file_path)
        self.reader.streams = self.__streams['json'] + self.__streams['default']
        return self.reader.parse()


def is_an_xml(xml_file):
    """
    Checks to see if XML file is a legit csv file

    :param xml_file: XML File path
    :return: True/False
    """

    if not exists(xml_file):
        raise Exception("'xml_file' %r does not exist" % xml_file)

    try:
        xml_parse(xml_file)
        return True
    except:
        pass

    return False


def unique_id():
    return get_random_string(length=16,
                             allowed_chars=u'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789')
